;;; erc-pcomplete.el --- Provides programmable completion for ERC

;; Copyright (C) 2002  Free Software Foundation, Inc.

;; Author: Sacha Chua <sacha@free.net.ph>
;; Keywords: comm, convenience
;; URL: http://www.emacswiki.org/cgi-bin/wiki.pl?ErcCompletion

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; This file replaces erc-complete.el.  It provides nick completion
;; for ERC based on pcomplete.  If you do not have pcomplete, you may
;; try to to erc-complete.el.
;;
;; To use, (require 'erc-auto) or (require 'erc-pcomplete), then
;;   (erc-pcomplete-mode 1)
;;
;; If you want erc-pcomplete to list the most recent speakers first,
;; add the following line to your .emacs after loading the erc-pcomplete
;; module.
;;
;;   (add-hook 'erc-server-PRIVMSG-hook
;;             'erc-pcomplete-ordering-PRIVMSG-handler)
;;
;; See CREDITS for other contributors.
;;
;;; Code:

(require 'pcomplete)
(require 'erc)
(require 'erc-compat)

(defconst erc-pcomplete-version "Revision: 1.22"
  "ERC pcomplete mode revision")

(defcustom erc-pcomplete-nick-postfix ": "
  "*When `pcomplete' is used in the first word after the prompt,
add this string to nicks completed."
  :group 'erc
  :type 'string)

;;;###autoload (autoload 'erc-completion-mode "erc-pcomplete" nil t)
(define-erc-module pcomplete Completion
  "In ERC Completion mode, the TAB key does completion whenever possible."
  ((add-hook 'erc-mode-hook 'pcomplete-erc-setup)
   (add-hook 'erc-complete-functions 'erc-pcomplete)
   (erc-buffer-list #'pcomplete-erc-setup))
  ((remove-hook 'erc-mode-hook 'pcomplete-erc-setup)
   (remove-hook 'erc-complete-functions 'erc-pcomplete)))

(defun erc-pcomplete ()
  "Complete the nick before point."
  (interactive)
  (when (> (point) (erc-beg-of-input-line))
    (let ((last-command (if (eq last-command 'erc-complete-word)
			    'pcomplete
			  last-command)))
      (call-interactively 'pcomplete))
    t))

;;; Setup function

(defun pcomplete-erc-setup ()
  "Setup `erc-mode' to use pcomplete."
  (set (make-local-variable 'pcomplete-ignore-case)
       t)
  (set (make-local-variable 'pcomplete-use-paring)
       nil)
  (set (make-local-variable 'pcomplete-suffix-list)
       '(?  ?:))
  (set (make-local-variable 'pcomplete-parse-arguments-function)
       'pcomplete-parse-erc-arguments)
  (set (make-local-variable 'pcomplete-command-completion-function)
       'pcomplete/erc-mode/complete-command)
  (set (make-local-variable 'pcomplete-command-name-function)
       'pcomplete-erc-command-name)
  (set (make-local-variable 'pcomplete-default-completion-function)
       (lambda () (pcomplete-here (pcomplete-erc-nicks)))))

;;; Programmable completion logic

(defun pcomplete/erc-mode/complete-command ()
  (pcomplete-here
   (append
    (pcomplete-erc-commands)
    (pcomplete-erc-nicks erc-pcomplete-nick-postfix))))

(defun pcomplete/erc-mode/CLEARTOPIC ()
  (pcomplete-here (pcomplete-erc-channels)))

(defun pcomplete/erc-mode/DEOP ()
  (while (pcomplete-here (pcomplete-erc-ops))))

(defun pcomplete/erc-mode/DESCRIBE ()
  (pcomplete-here (pcomplete-erc-nicks)))

(defun pcomplete/erc-mode/IDLE ()
  (while (pcomplete-here (pcomplete-erc-nicks))))

(defun pcomplete/erc-mode/KICK ()
  (pcomplete-here (pcomplete-erc-channels))
  (pcomplete-here (pcomplete-erc-nicks)))

(defun pcomplete/erc-mode/LOAD ()
  (pcomplete-here (pcomplete-entries)))

(defun pcomplete/erc-mode/MODE ()
  (pcomplete-here (pcomplete-erc-channels))
  (while (pcomplete-here (pcomplete-erc-nicks))))

(defun pcomplete/erc-mode/ME ()
  (while (pcomplete-here (pcomplete-erc-nicks))))

(defun pcomplete/erc-mode/SAY ()
  (pcomplete-here (pcomplete-erc-nicks))
  (pcomplete-here (pcomplete-erc-nicks))
  (while (pcomplete-here (pcomplete-erc-nicks))))

(defun pcomplete/erc-mode/MSG ()
  (pcomplete-here (append (pcomplete-erc-all-nicks)
			  (pcomplete-erc-channels)))
  (while (pcomplete-here (pcomplete-erc-nicks))))

(defun pcomplete/erc-mode/NAMES ()
  (while (pcomplete-here (pcomplete-erc-channels))))

(defalias 'pcomplete/erc-mode/NOTICE 'pcomplete/erc-mode/MSG)

(defun pcomplete/erc-mode/OP ()
  (while (pcomplete-here (pcomplete-erc-not-ops))))

(defun pcomplete/erc-mode/PART ()
  (pcomplete-here (pcomplete-erc-channels)))

(defalias 'pcomplete/erc-mode/LEAVE 'pcomplete/erc-mode/PART)

(defun pcomplete/erc-mode/QUERY ()
  (pcomplete-here (append (pcomplete-erc-all-nicks)
			  (pcomplete-erc-channels)))
  (while (pcomplete-here (pcomplete-erc-nicks)))
  )

(defun pcomplete/erc-mode/SOUND ()
  (while (pcomplete-here (pcomplete-entries))))

(defun pcomplete/erc-mode/TOPIC ()
  (pcomplete-here (pcomplete-erc-channels)))

(defun pcomplete/erc-mode/WHOIS ()
  (while (pcomplete-here (pcomplete-erc-nicks))))

(defun pcomplete/erc-mode/UNIGNORE ()
  (pcomplete-here (with-current-buffer (erc-server-buffer) erc-ignore-list)))

;;; Functions that provide possible completions.

(defun pcomplete-erc-commands ()
  "Returns a list of strings of the defined user commands."
  (let ((case-fold-search nil))
    (mapcar (lambda (x)
	      (concat "/" (downcase (substring (symbol-name x) 8))))
	    (apropos-internal "erc-cmd-[A-Z]+"))))

(defun pcomplete-erc-ops ()
  "Returns a list of nicks with ops."
  (mapcar (lambda (x)
	    (if (car (cdr x)) (car x))) channel-members))

(defun pcomplete-erc-not-ops ()
  "Returns a list of nicks without ops."
  (mapcar (lambda (x)
	    (if (not (car (cdr x))) (car x))) channel-members))

(defun pcomplete-erc-nicks (&optional postfix)
  "Returns a list of nicks in the current channel."
  (mapcar (lambda (x)
	    (if (car x)
		(concat (car x) postfix)))
	  channel-members))

(defun pcomplete-erc-all-nicks (&optional postfix)
  "Returns a list of all nicks on the current current server."
  (let (res)
    (mapc
     (lambda (buf)
       (with-current-buffer buf
	 (mapc (lambda (elt)
                 (if (car elt)
		     (add-to-list 'res (concat (car elt) postfix) res)))
               channel-members)))
     (erc-channel-list erc-process))
    res))

(defun pcomplete-erc-channels ()
  "Returns a list of channels associated with the current server."
  (mapcar (lambda (buf) (with-current-buffer buf (erc-default-target)))
	  (erc-channel-list erc-process)))

;;; Functions for parsing

(defun pcomplete-erc-command-name ()
  "Returns the command name of the first argument."
  (if (eq (elt (pcomplete-arg 'first) 0) ?/)
      (upcase (substring (pcomplete-arg 'first) 1))
    "SAY"))

(defun pcomplete-parse-erc-arguments ()
  "Returns a list of parsed whitespace-separated arguments.
These are the words from the beginning of the line after the prompt
up to where point is right now."
  (let* ((start (save-excursion (erc-bol) (point)))
	 (end (point))
	 args beginnings)
    (save-excursion
      (if (< (skip-chars-backward " \t\n" start) 0)
	  (setq args '("")
		beginnings (list end)))
      (setq end (point))
      (while (< (skip-chars-backward "^ \t\n" start) 0)
	(setq beginnings (cons (point) beginnings)
	      args (cons (buffer-substring-no-properties
			  (point) end)
			 args))
	(skip-chars-backward " \t\n" start)
	(setq end (point))))
    (cons args beginnings)))

;;; Functions to order nick completions such that the most recent
;;; speakers are listed first.

(defun erc-pcomplete-bring-channel-member-to-front (nick)
  "Brings the entry `channel-members' for NICK to the front of the list.
This function should be called with the current buffer set to the
appropriate channel buffer.
See also `erc-pcomplete-ordering-PRIVMSG-handler'."
  (setq nick (erc-downcase nick))
  (let* ((names-list (erc-get-current-channel-members))
         (names names-list)
         temp-var)
    (when (and names
               (not (string= nick (erc-downcase (caar names)))))
      (catch 'done
        (while (cdr names)
          (if (string= nick (erc-downcase (caadr names)))
              (throw 'done nil)
            (setq names (cdr names)))))
      (when (cdr names)
        (setq channel-members (cons (cadr names) names-list))
        (setcdr names (cddr names))))))

(defun erc-pcomplete-ordering-PRIVMSG-handler (proc parsed)
  "Calls `erc-pcomplete-bring-channel-member-to-front' with the
nickname of the sender, if this PRIVMSG is an unignored channel
message.  Always returns `nil' so that it can conveniently be
used in server message hooks.  If this is added to
`erc-server-PRIVMSG-hook', nickname completions will be ordered
such that the most recent speakers are listed first."
  (let ((sender-spec (aref parsed 1))
        (cmd (aref parsed 0))
        (tgt (aref parsed 2))
        (msg (aref parsed 3)))
    (if (and (not (or (erc-ignored-user-p sender-spec)
                      (erc-ignored-reply-p msg tgt proc)))
             (erc-channel-p tgt))
        (let* ((sndr (erc-parse-user sender-spec))
               (nick (car sndr))
               (buffer  (erc-get-buffer tgt proc)))
          (if buffer
              (with-current-buffer buffer
                (erc-pcomplete-bring-channel-member-to-front nick))))))
  nil)

(provide 'erc-pcomplete)

;;; erc-pcomplete.el ends here
